/* ############################################################
 * Programm zum Verstecken von Daten im ppm Image
 * Autor: Helmut Fahrion
 * E-Mail: hefa@gmx.net
 * Lizenz: GPL
 * Datum: 09.2004
 * ############################################################
 */

// fr crypt
#define _XOPEN_SOURCE
#include <unistd.h>

#define _GNU_SOURCE
#include <getopt.h>		// long options
#include <stdio.h>		// printf
#include <string.h>		// memcpy
#include <errno.h>		// error
#include <sys/mman.h>		// mmap

#define __USE_XOPEN_EXTENDED	// fr Random
#include <stdlib.h>		// getenv

// netpbm tools, functions to read and write PAM or PNM image files
#include <pnm.h>
// Zentrale Versionsnummer
#include "stego-version.h"

// Daten des Bildes
typedef struct
{
  int rows, cols;
  xel **array;
} bild_t;

bild_t bild;			// Bilddaten

// Daten der Nachricht
typedef struct
{
  char *crray;			// Array fr die Bits
  int *rndarray;		// Array fr Zufallszahlen
  unsigned char state[256];	// Array fr Zufallsverteilung
  unsigned int calc[256];	// Array fr Zufallsverteilung
  // Dateizeiger auf das File welches versteckt werden soll.
  char *nachricht;
} array_t;

array_t array;			// Cryptdaten


// +++++ Mappt das Nachrichtenfile in den Hauptspeicher +++++
void
mmap_file (char **data, int *size, const char *name)
{
  int fd;
  struct stat fs;
  char *p;

  if ((fd = open (name, O_RDONLY, 0)) == -1)
    {
      fprintf (stderr, "Kann %s nicht ffnen!\n", name);
      exit (1);
    }
  if (fstat (fd, &fs) == -1)
    {
      perror ("fstat");
      exit (1);
    }

  if ((p =
       mmap (NULL, fs.st_size, PROT_READ, MAP_SHARED, fd, 0)) == MAP_FAILED)
    {
      perror ("mmap");
      exit (1);
    }

  close (fd);

  fprintf (stderr, "mmap() 0x%04X Bytes\n", (unsigned int) fs.st_size);

  *data = p;
  *size = fs.st_size;
}


// +++ vor Programmende aufrumen +++
void
make_free (int nsize)
{
  // immer schn zumachen
  if (array.nachricht)
    munmap (array.nachricht, nsize);

  // Kontrollarray
  if (array.crray)
    free (array.crray);

  // Zufallsverteilung  
  if (array.rndarray)
    free (array.rndarray);

  pnm_freearray (bild.array, bild.rows);
}

// Liest die Bilddaten ein
// Eingabe: nsize   zur Grssenbestimmung
// Ausgabe: maxval  Grsste Wert des Tupel (r g b)
// Ausgabe: format  Fileformat
void
make_eingabe (xelval * maxval, int *format, int nsize, int orig, int seek)
{
  xel p;
  int x, y, p1, p2, anz, maxc;
  unsigned int alles;

  bild.array = pnm_readpnm (stdin, &bild.cols, &bild.rows, maxval, format);

  // Kontrollarray damit kein Pixel doppelt gesetzt wird
  anz = bild.cols * bild.rows;
  array.crray = malloc (sizeof (char) * anz);
  memset (array.crray, 0, sizeof (char) * anz);

  // Kontrollarray damit kein Pixel doppelt gesetzt wird
  array.rndarray = malloc (sizeof (int) * anz);
  // vorinitialisert
  for (x = 0; x < anz; x++)
    {
      array.rndarray[x] = x;
    }

  // Teste ob alles reinpasst
  alles = 4;			// 4Caracters fr die Gre 
  alles += 22;			// 22 Bytes fr die MD5 Passwortsumme 
  alles += nsize;		// Bytes der Nachricht

  // wir brauchen fr jedes Bit ein Byte
  alles *= 8;

  // Passt alles rein ?
  if ((unsigned) (bild.rows * bild.cols) >= alles)
    {
      // Es hat geklappt!
      fprintf (stderr,
	       "Es sind %d Bits zu kodieren! %d Bytes sind vorhanden.\n",
	       alles, (bild.rows * bild.cols));
    }
  else
    {
      fprintf (stderr,
	       "Nicht gengend Bytes fr die zu codierenden Bytes vorhanden.\nEs sind %d Bits zu kodieren! %d Bytes sind vorhanden.\n",
	       alles, (bild.rows * bild.cols));

      // Programm beenden 
      make_free (nsize);
      exit (1);
    }

  if (bild.array && array.crray)
    {
      fprintf (stderr, "cols:%d rows:%d maxval:%u format:0x%x \n",
	       bild.cols, bild.rows, (unsigned int) *maxval, (*format));
    }
  else
    {
      fprintf (stderr, "konnte tuplearray nicht allocen!\n");
    }

  // initialisiere das Array mit "zufallszahlen" fr die 
  // Verteilung der Bits
  if (array.rndarray)
    {
      // Mache das Verteilungsarray, aus einem Histogramm
      if (orig)
	{
	  // Std Zufallszahlenfolge
	  srandom (seek);
	  fprintf (stderr, "Init Zufallszahlen mit seek().\n");
	}
      else
	{
	  memset (array.calc, 0, 256);
	  for (maxc = x = 0; x < anz; x++)
	    {
	      // hohle Pixel
	      p = bild.array[x / bild.cols][x % bild.cols];

	      // Pixelwert
	      /*
	         y = (int) (((p.r & 0xfe) >> 1) +
	         ((p.g & 0xfe) >> 1) +
	         ((p.b & 0xfe) >> 1) / 3.0);
	       */
	      y = (int) (((p.r & 0xfe) >> 1) +
			 ((p.g & 0xfe) >> 1) + ((p.b & 0xfe) >> 1));

	      if (y > maxc)
		maxc = y;

	      // Die Position ist dem Wert entsprechend
	      // ++ gibt die Anzahl
	      array.calc[y % 256]++;

	    }

	  for (x = 0; x < 256; x++)
	    {
	      array.state[x] = (unsigned char)
		(((double) array.calc[x] / (double) maxc) * 255.0);
	    }

	  // Ausgabe der initstart Arrays
	  fprintf (stderr, "Init Zufallszahlen mit Histogramm.\n");
	  /*
	     for (x = 0; x < 256; x++)
	     {
	     fprintf (stderr, "%u:", array.calc[x]);
	     fprintf (stderr, "%u ", array.state[x]);
	     }
	     fprintf (stderr, "\n");
	   */
	  initstate (seek, array.state, 256);
	  setstate (array.state);
	}

      // Zahlen werden durcheinandergewrfelt
      for (x = 0; x < anz; x++)
	{
	  // 2 Pixel zum tauschen
	  //p1=(int) ((double)(anz) * ((double)random() / (RAND_MAX+1.0)));
	  p1 = (x % anz);
	  p2 =
	    (int) ((double) (anz) * ((double) random () / (RAND_MAX + 1.0)));

	  // tauschen
	  y = array.rndarray[p1];
	  array.rndarray[p1] = array.rndarray[p2];
	  array.rndarray[p2] = y;
	}

    }
  else
    {
      fprintf (stderr, "konnte randarray nicht allocen!\n");
    }
}

// +++ die Geringfgkeit der nderung kann auch auffllig sein.
// Daher wird versucht die nderung "einzupassen".
// Also die Werte werden immer etwas gemittelt...
int
mpixel_sub (int *aminus, int *aplus, int *apass, int a, int l, int r)
{
  int dl, dr;

  // Anpassung, Links Rechts
  if ((a > l && a > r) || (a > l && a == r) || (a == l && a > r))
    {
      (*aminus)++;
      (*apass)++;
      return (a - 1);
    }
  if ((a < l && a == r) || (a < l && a < r) || (a == l && a < r))
    {
      (*aplus)++;
      (*apass)++;
      return (a + 1);
    }
  if ((a < l && a > r) || (a > l && a < r))
    {
      dl = l - a;
      dr = a - r;

      if (dr > dl)		// Werte werden gemittelt
	{
	  (*aplus)++;
	  (*apass)++;
	  return (a + 1);	// Grer oder kleiner beides geht
	}
      else
	{
	  (*aminus)++;
	  (*apass)++;
	  return (a - 1);
	}
    }
  // Sind alle Varianten, nur (a==l && a==r) fehlt das ist 
  // zwangssetzen also, nix gefunden !  
  return 256;
}

// Testet die Pixel, ob der Umgebung entsprechend erhht oder erniedrigt werden
// kann.
int
mpixel (int *aminus, int *aplus, int *apass, int *m)
{
  int x, y;

  for (y = 1; y < 7; y += 2)
    {
      // Teste die Pixel in der Reihenfolge  links-rechts, oben-unten,
      // linksoben-rechtsunten, rechtsoben-linksunten
      x = mpixel_sub (aminus, aplus, apass, m[0], m[y], m[y + 1]);
      if (x != 256)
	return x;
    }

  return 256;
}

// Pixel ndern auch wenn es zu keiner Verbesserung des Bildes fhrt
int
zwang_pixel (int *aminus, int *aplus, int a, int mitte)
{
  // Ansonsten halt Zwang
  if (mitte < a)
    {
      (*aminus)++;
      return (a - 1);
    }
  else
    {
      (*aplus)++;
      return (a + 1);
    }
}

// trennt die Farbwerte aus dem Pixelarray
void
mk_rot (int *m, xel * p)
{
  int x;
  for (x = 0; x < 9; x++)
    m[x] = p[x].r;
}

void
mk_gruen (int *m, xel * p)
{
  int x;
  for (x = 0; x < 9; x++)
    m[x] = p[x].g;
}

void
mk_blau (int *m, xel * p)
{
  int x;
  for (x = 0; x < 9; x++)
    m[x] = p[x].b;
}

// Reihenfolgenderung beim Pixelsetzen, r, g oder b ! Es soll sich halt immer
// abwechseln.
int
reihenfolge (xel * r, int *aminus, int *aplus, int *apass, xel * p,
	     int reihenf)
{
  int matrix[9], m;
  static int letzte;
  *r = p[0];
  m = 0;
  switch (reihenf)
    {
    case 0:
      mk_blau (matrix, p);
      r->b = mpixel (aminus, aplus, apass, matrix);
      letzte = reihenf;		// merke mir die letzte farbe
      if (r->b == 256)
	return 0;		// nicht fertig, nocheins probieren
      break;

    case 1:
      mk_rot (matrix, p);
      r->r = mpixel (aminus, aplus, apass, matrix);
      letzte = reihenf;		// merke mir die letzte farbe
      if (r->r == 256)
	return 0;		// nicht fertig, nocheins probieren
      break;

    case 2:
      mk_gruen (matrix, p);
      r->g = mpixel (aminus, aplus, apass, matrix);
      letzte = reihenf;		// merke mir die letzte farbe
      if (r->g == 256)
	return 0;		// nicht fertig, nocheins probieren
      break;

      // Wenn bis hier, dann setzen
    case 3:
      // fprintf (stderr, "Zwangpixel vor:  r:%d g:%d b:%d\n", r->r, r->g, r->b);

      // Suche die "innere Mitte"
      switch (letzte)
	{
	case 0:
	  m = (p[0].r + p[0].g) / 2;
	  r->b = zwang_pixel (aminus, aplus, p[0].b, m);
	  break;
	case 1:
	  m = (p[0].g + p[0].b) / 2;
	  r->r = zwang_pixel (aminus, aplus, p[0].r, m);
	  break;
	case 2:
	  m = (p[0].r + p[0].b) / 2;
	  r->g = zwang_pixel (aminus, aplus, p[0].g, m);
	  break;
	}
      //fprintf (stderr, "Zwangpixel nach: r:%d g:%d b:%d m:%d\n", r->r, r->g, r->b, m);

      // die "innere" Anpassung ist nicht so schlecht, sollte gut sein!
      if (m > 2)
	(*apass)++;
      break;

    default:
      break;
    }

  return 1;			// Fertig
}

// immer mal ein anderen Farbwert ndern
xel
umschalten (int *aminus, int *aplus, int *apass, xel * p, char showmode)
{
  int reihf[3], x;

  xel r = p[0];

  // +++++  Testzweck um die Verteilung zu sehen +++++++++
  if (showmode)
    {
      if (((p[0].r + p[0].g + p[0].b) & 1))
	{
	  (*aminus)++;
	  r.g = 254;
	  r.r = r.b = 0;
	}
      else
	{
	  (*aplus)++;
	  r.r = 255;
	  r.b = r.g = 0;
	}
    }
  else
    {
      // Immer ein anderer Farbwert ist drann, die Regelmigkeit
      // knnte auch auffllig sein!
      // rgb gbr brg
      switch (random () % 3)
	{
	case 0:
	  reihf[0] = 0;
	  reihf[1] = 1;
	  reihf[2] = 2;
	  break;
	case 1:
	  reihf[0] = 1;
	  reihf[1] = 2;
	  reihf[2] = 0;
	  break;
	case 2:
	  reihf[0] = 2;
	  reihf[1] = 0;
	  reihf[2] = 1;
	  break;
	default:
	  break;
	}

      reihf[3] = 3;

      // Teste jedesmal immer anderer Reihenfilge,
      // 3 ist Zwangssetzen
      for (x = 0; x < 4; x++)
	if (reihenfolge (&r, aminus, aplus, apass, p, reihf[x]))
	  return r;
    }

  return r;
}

// Bit je Pixel lesen
void
get_pixel (unsigned int *sumcnt, int h)
{
  xel p;

  // umrechnen in rows und cols
  p = bild.array[h / bild.cols][h % bild.cols];

  // crray ist nicht rekursiv, sondern die richtige
  // Reihenfolge
  array.crray[(*sumcnt)++] = ((p.r + p.g + p.b) & 1);
}

// Testet ob die Pixel umgeschaltet werden mssen oder
// nicht.
void
chk_umschalten (int *aminus, int *aplus, int *apass,
		int h, char b, char showmode)
{
  int r, c, hx;
  xel xela[9];

  // Umrechnen in rows und cols
  r = h / bild.cols;
  c = h % bild.cols;

  xela[0] = bild.array[r][c];	// Pixel 

  // linke Pixel
  hx = ((h - 1) >= 0) ? h - 1 : h;
  xela[1] = bild.array[hx / bild.cols][hx % bild.cols];

  // rechte Pixel
  hx = ((h + 1) <= (bild.rows * bild.cols - 1)) ? h + 1 : h;
  xela[2] = bild.array[hx / bild.cols][hx % bild.cols];

  // obere Pixel
  hx = ((h - bild.cols) >= 0) ? h - bild.cols : h;
  xela[3] = bild.array[hx / bild.cols][hx % bild.cols];

  // untere Pixel
  hx =
    ((h + bild.cols) <=
     (bild.rows * bild.cols - bild.cols)) ? h + bild.cols : h;
  xela[4] = bild.array[hx / bild.cols][hx % bild.cols];

  // obere linke Pixel
  hx = ((h - bild.cols - 1) >= 0) ? h - bild.cols - 1 : h;
  xela[5] = bild.array[hx / bild.cols][hx % bild.cols];

  // untere rechte Pixel
  hx =
    ((h + bild.cols + 1) <=
     (bild.rows * bild.cols - bild.cols)) ? h + bild.cols + 1 : h;
  xela[6] = bild.array[hx / bild.cols][hx % bild.cols];

  // obere rechte Pixel
  hx = ((h - bild.cols + 1) >= 0) ? h - bild.cols + 1 : h;
  xela[7] = bild.array[hx / bild.cols][hx % bild.cols];

  // untere linke Pixel
  hx =
    ((h + bild.cols - 1) <=
     (bild.rows * bild.cols - bild.cols)) ? h + bild.cols - 1 : h;
  xela[8] = bild.array[hx / bild.cols][hx % bild.cols];

  // Je nach Bit entscheiden, Passt oder nicht
  if (b)
    {
      if (((xela[0].r + xela[0].g + xela[0].b) & 1) == 0)
	{
	  //  fprintf (stderr, "0");
	  bild.array[r][c] = umschalten (aminus, aplus, apass, xela,
					 showmode);
	}
    }
  else
    {
      if (((xela[0].r + xela[0].g + xela[0].b) & 1) == 1)
	{
	  // fprintf (stderr, "1");
	  bild.array[r][c] = umschalten (aminus, aplus, apass, xela,
					 showmode);
	}
    }

}

void
ausgabe (char c)
{
  fprintf (stderr, "%c", c);
}

// Pixel je Bit schreiben
//
// Ausgabe:
//
//
// Eingabe:
//
int
make_pixel (int *aplus, int *aminus, int *apass,
	    unsigned int *sumcnt, unsigned char *sumbit,
	    const char *versteck,
	    int x, const char *cryptkey, int verxor, int nsize, char showmode)
{
  char b;
  int h;

  h = array.rndarray[x];

  // Kontrollarray setzen
  if (array.crray[h])
    {
      fprintf (stderr, "Fehler: Pixel doppelt gesetzt!!!\n");
      return 1;			// fertig, da Fehler
    }
  else
    array.crray[h] = 1;

  // Umschaltung einbauen 
  // +++ MD5 Summe +++
  if ((*sumcnt) < 22)
    {
      // +++ Verschleierung der md5 Summe +++
      b = (cryptkey[(*sumcnt)] + versteck[(*sumcnt)]) % 256;
      b >>= (*sumbit);		// Shift frs Bit damit es ganz rechts ist
      b &= 1;			// alle Wegbbeln, auer das letzte

      // Zhler erhhen fr das nchste mal
      (*sumbit)++;
      if ((*sumbit) == 8)
	{
	  (*sumcnt)++;
	  (*sumbit) = 0;
	}
      // nur jedes 10. ausgeben
      if (!(x % 10))
	ausgabe ('s');
    }
  else
    {
      // +++ Schreibe die Anzahl der Bytes der Nachricht +++
      if ((*sumcnt) == 22)
	{
	  // wird Passwortabhngig verxodert
	  b = ((nsize ^ verxor) >> (*sumbit)) & 1;

	  // Zhler erhhen fr das nchste mal,
	  (*sumbit)++;
	  if ((*sumbit) == 32)
	    {
	      (*sumcnt)++;
	      (*sumbit) = 0;
	    }

	  if (!(x % 10))
	    ausgabe ('b');
	}
      else
	{
	  if (((*sumcnt) - 23) < (unsigned) nsize)
	    {
	      // +++ Hole Byte aus der gemappten Datei,
	      // wird passwortabhngig verxodert
	      // dann entsprechend des gesuchten Bits geschiftet
	      // und mit 1 "gebgelt" so bleibt das Bit brig ! +++
	      b = ((array.nachricht[*sumcnt - 23] ^
		    versteck[(*sumcnt - 23) % 22]) >> (*sumbit)) & 1;

	      // jetzt die Nachricht schreiben 
	      // Zhler erhhen fr das nchste mal
	      (*sumbit)++;
	      if ((*sumbit) == 8)
		{
		  (*sumcnt)++;
		  (*sumbit) = 0;
		}
	      if (!(x % 100))
		ausgabe ('p');
	    }
	  else
	    {
	      // fertig !!!
	      return 1;
	    }
	}
    }

  // schreibt die Pixel falls ntig
  chk_umschalten (aminus, aplus, apass, h, b, showmode);

  return 0;
}

// Gehe durch das Array baue Pixel auf
void
make_stego_put (int *aplus, int *aminus, int *apass,
		int end, const char *cryptkey, char showmode, int verxor,
		int nsize, const char *versteck)
{
  int x;
  unsigned int sumcnt;
  unsigned char sumbit;

  for (sumcnt = sumbit = x = 0; x <= end; x++)
    {
      // hier erfolgt das ummappen auf die "Zufallsverteilung"
      if (make_pixel (aplus, aminus, apass, &sumcnt, &sumbit,
		      versteck, x, cryptkey, verxor, nsize, showmode))
	{
	  return;
	}
    }
}

// Gehe durch das Array und lese die Pixel
void
make_stego_get (int end)
{
  int x;
  unsigned int sumcnt;

  for (sumcnt = x = 0; x <= end; x++)
    {
      get_pixel (&sumcnt, array.rndarray[x]);
    }
}

// Das Bild schreiben
void
make_ausgabe_bild (xelval maxval, int format)
{
  // ASCII oder Binrmode 
  //char forceplain = 1;
  fprintf (stderr, "\nAusgabe: cols:%d rows:%d maxval:%d format:0x%x \n",
	   bild.cols, bild.rows, maxval, format);

  // Berechnete Array ausgeben, 1 ASCII
  pnm_writepnm (stdout, bild.array, bild.cols, bild.rows, maxval, format, 1);
}


// Widergefundene Nachricht ausgeben
void
make_ausgabe_nachricht (const char *cryptkey, int verxor,
			const char *versteck)
{
  char c, b;
  char md5[23];
  int x, bytes;
  unsigned int sumcnt;

  // Zhler erhhen fr das nchste mal
  memset (md5, 0, 23);

  // 22 Bytes der MD5 Sum lesen
  for (x = 0; x < 22; x++)
    {
      //b = 0;
      for (b = c = 0; c < 8; c++)
	{
	  b |= ((array.crray[x * 8 + c]) << c);
	}

      md5[x] = (b - versteck[x]) % 256;
      //fprintf (stderr, "m");
    }
  sumcnt = x;

  fprintf (stderr, "md5: %s\n", md5);
  fprintf (stderr, "cryptkey: %s\n", cryptkey);

  // Passwortvergleich
  if (strcmp (md5, cryptkey))
    {
      fprintf (stderr, "Passwort falsch oder keine Daten!\n");
      exit (1);
    }

  fprintf (stderr, "Passwort OK! Daten gefunden!\n");

  // ++++ jetzt weiterlesen, Bytes +++++
  for (bytes = x = 0; x < 32; x++)
    {
      bytes |= ((array.crray[sumcnt * 8 + x]) << x);
      fprintf (stderr, "b");
    }
  // counter im array
  sumcnt += 4;

  // wird passwortabhngig verxodert
  bytes ^= verxor;

  fprintf (stderr, " Anzahl Bytes=%d\n", bytes);

  // ++++ jetzt weiterlesen, die Nachricht +++++
  for (x = 0; x < bytes; x++)
    {
      for (b = c = 0; c < 8; c++)
	b |= ((array.crray[(sumcnt + x) * 8 + c]) << c);

      b ^= versteck[x % 22];

      // +++ Ausgabe:  :-) +++
      printf ("%c", b);
    }
}


// *********** main ***********************
int
main (int argc, char *argv[])
{
// Der Hilfetext
  const char usage[] =
    "%s: Stego (Steganographie Tool) %s\n"
    "Lizenze: GPL\n"
    "Notation: %s [--version|-v] [--help|--hilfe|-h] [--key|-k password]\n"
    "   [--env|-e] [--nachricht|-n <dateiname>] [--show|-s] [--read|-r]\n\n"
    "Bitte lies die README und tippe einfach:\n"
    ">man stego\n" "\nViel Spass...\n";

  static const struct option long_options[] = {
    {"env", 1, 0, 'e'},
    {"help", 0, 0, 'h'},
    {"input", 1, 0, 'i'},
    {"key", 1, 0, 'k'},
    {"nachricht", 1, 0, 'n'},
    {"orig", 1, 0, 'o'},
    {"read", 0, 0, 'r'},
    {"show", 0, 0, 's'},
    {"version", 0, 0, 'v'},
    {"hilfe", 0, 0, '?'},
    {0, 0, 0, 0}
  };

  int opt;			// Kommandozeilenoption parsen
  char *ck;			// Zeiger auf Crypt String
#define pwd_maxlen 1024		// Maximale Lnge es Passwortes
  char pwd[pwd_maxlen];		// Eingangs Passwort
  FILE *pwd_stream;		// Name des Passwortfiles

  // Schlssel um die Daten "wiederzufinden"!
  char cryptkey[30] = "123456789012345678901234567890";
  // Schlssel um die Daten zu "verstecken"!
  // ist ein Vigenre-Verfahren, ist fr ein Kryptanalytiker
  // Spielkram, der Text ist aber schon verschlsselt. Diese
  // Verschlsselung dient nur der "Verscheierung" um eventuellen
  // Statistischen Analysen zu entzeihen.
  char versteck[30] = "123456789012345678901234567890";

  int showmode;			// Zeigt die Vernderten Pixel an, fr Kontrollzwecke
  int verxor;			// Wert zum Verstecken der MD5 Summe
  int stegmode;			// Lesen oder Schreiben
  int format;			// Fileformat
  int nsize;			// Gre der Nachricht
  int origrand;			// Verwende die Originale Zufallszahlenfolge
  int option_index;
  int seek;			// Wert fr den Zufallszahlengenerator

  xelval maxval;		// Maxvalue der Pixel

  array.nachricht = NULL;
  pnm_init (&argc, argv);

  origrand = nsize = showmode = stegmode = verxor = option_index = 0;

  // parameter auswerten  "vhsrk:n:"
  while ((opt =
	  getopt_long (argc, argv, "h?rsvoei:k:n:", long_options,
		       &option_index)) != EOF)
    {
      switch (opt)
	{
	case 'o':
	  origrand = 1;
	  break;
	case 's':
	  showmode = 1;
	  break;
	case 'r':
	  stegmode = 1;
	  break;
	case 'n':
	  mmap_file (&array.nachricht, &nsize, optarg);
	  break;
	  // Der Key liegt in einer Datei
	case 'i':
	  pwd_stream = fopen (optarg, "r");
	  if (!pwd_stream)
	    {
	      perror ("Fehler beim ffnen der Key-Datei!\n");
	      exit (1);
	    }
	  // Ganze Datei ein String
	  fgets (pwd, pwd_maxlen, pwd_stream);
	  //fprintf (stderr, "PWD aus Datei gelesen! \"%s\"\n", pwd);
	  fprintf (stderr, "PWD aus Datei gelesen!\n");
	  fclose (pwd_stream);	// immer schn wieder zu machen
	  goto cryptpwd;
	  break;
	  // macht ein MD5 Passwort, mit dem wir erkennen ob die Daten im File
	  // sind
	case 'e':
	  strncpy (pwd, getenv ("STEGOPWD"), 99);
	  goto cryptpwd;
	  break;

	case 'k':
	  //memset (pwd, 0, 100);
	  strncpy (pwd, optarg, 99);
	  // +++ :-) +++  
	cryptpwd:
	  if (!(ck = crypt (pwd, "$1$hefasteg$")))
	    {
	      fprintf (stderr, "# Fehler bei der Keybildung!\n");
	    }
	  strncpy (cryptkey, ck + 12, 23);

	  // Bilde in Abhngigkeit vom Passwort den Versteckkey !
	  // cryptkey darf man ja nicht nehmen denn er ist in der Datei
	  // gespeichert!

	  if (!(ck = crypt (pwd, "$1$versteck$")))
	    {
	      fprintf (stderr, "# Fehler bei der Key2 Bildung!\n");
	    }
	  strncpy (versteck, ck + 12, 23);

	  //  Ein Passwortabhngiges XOR zusammenbasteln
	  /*
	     verxor =
	     (versteck[1] << 0) |
	     (versteck[2] << 7) |
	     (versteck[3] << (7*2)) |
	     (versteck[4] << (7*3)) |
	     (versteck[5] << (7*4)); */
	  verxor =
	    ((versteck[1] + versteck[2]) << 0) |
	    ((versteck[3] + versteck[4]) << 8) |
	    ((versteck[5] + versteck[6]) << (8 * 2)) |
	    ((versteck[7] + versteck[8]) << (8 * 3));

	  fprintf (stderr, "berechneter key: %s pwd:%s\n", cryptkey, pwd);
	  fprintf (stderr, "berechneter key2: %s\n", versteck);
	  fprintf (stderr, "verxor: %x\n", verxor);
	  break;
	case '?':
	case 'h':
	  printf (usage, argv[0], version, argv[0]);
	  return 0;
	  break;
	case 'v':
	  printf ("Version: %s\n", version);
	  return 0;
	  break;
	default:
	  printf ("?? getopt returned character code 0%o ??\n", opt);
	  return 1;
	  break;
	}
    }

  // Bischen Eingabeberprfung
  if (strlen (pwd) < 5)
    {
      fprintf (stderr, "Passwortlnge zu klein!\n");
      return 1;
    }

  if (!stegmode)
    {
      if (!nsize)
	{
	  // keine Nachricht aber trotzdem kodieren ? was soll das?
	  fprintf (stderr, "Keine Nachrichtendatei angegeben!\n");
	  fprintf (stderr, "Oder -r vergessen.\n");
	  return 1;
	}
    }

  // Zufallszahlen fr die Verteilung der Bits auf die RGB Werte
  // auch Passwortabhngig und Bildabhngig (row,col) !!!
  // , Pseudozufall sollte ausreichend sein
  // Es ist extrem Wichtig das die Verteilung der Information bei
  // Bild einlesen
  seek = ((versteck[9] + versteck[10]) << 0) |
    ((versteck[11] + versteck[12]) << 8) |
    ((versteck[13] + versteck[14]) << (8 * 2)) |
    ((versteck[15] + versteck[16]) << (8 * 3));

  make_eingabe (&maxval, &format, nsize, origrand, seek);

  // Nachricht einfgen oder lesen
  if (stegmode)
    {
      // lesen
      make_stego_get (bild.rows * bild.cols - 1);
      // Ausgabe
      make_ausgabe_nachricht (cryptkey, verxor, versteck);
    }
  else
    {				// Die Arbeit
      // Statistik
      int aplus, aminus, apass, sum, vershl;
      aplus = aminus = apass = sum = vershl = 0;
      double proz;

      // Gehe durch das Array, baue Pixel auf
      make_stego_put (&aplus, &aminus, &apass,
		      bild.rows * bild.cols - 1, cryptkey,
		      showmode, verxor, nsize, versteck);

      // Ausgabe, maxval und format bleibt, wie beim lesen
      make_ausgabe_bild (maxval, format);

      sum = aplus + aminus;
      vershl = abs (sum - apass);
      fprintf (stderr, "Statistik: +=%d -=%d sum=%d pass=%d schlecht=%d\n",
	       aplus, aminus, sum, apass, vershl);

      proz = ((double) vershl) / ((bild.rows * bild.cols));
      fprintf (stderr, "%lf%% der Pixel verschlechtert!\n", proz);
    }

  // vor beenden
  make_free (nsize);

  return 0;
}
